package application;

import infrastructure.DeterministicNodeIdFactory;
import infrastructure.DirectDeliveryNotification;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.util.Calendar;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import message.*;

import database.GossipAnalysisXMLDB;
import database.PartnerXMLDB;
import database.ProfileXMLDB;
import database.StatusXMLDB;
import database.TrafficAnalysisXMLDB;

import rice.Continuation;
import rice.Executable;
import rice.environment.Environment;
import rice.p2p.commonapi.Application;
import rice.p2p.commonapi.DeliveryNotification;
import rice.p2p.commonapi.Endpoint;
import rice.p2p.commonapi.Id;
import rice.p2p.commonapi.Message;
import rice.p2p.commonapi.Node;
import rice.p2p.commonapi.NodeHandle;
import rice.p2p.commonapi.RouteMessage;
import rice.p2p.commonapi.rawserialization.InputBuffer;
import rice.p2p.commonapi.rawserialization.MessageDeserializer;
import rice.p2p.util.BloomFilter;
import rice.pastry.PastryNode;
import rice.pastry.PastryNodeFactory;
import rice.pastry.socket.SocketPastryNodeFactory;

import toolbox.FixLengthQueue;
import toolbox.GossipLog;
import toolbox.HealthIndex;
import toolbox.Mailbox;
import toolbox.ObjectSerialize;
import toolbox.PartnerMap;
import toolbox.TrafficAnalysisRecorder;
import toolbox.UTPair;
import twclientapi.TwClientAPI;

public class Cuckoo implements Application, Runnable {
	private static final boolean DEBUG = false;
	
	//private static boolean BOOTSTRAP = false;
	
	private static final int BULK = 20;
	public static final int C = 5; //gossip constant
	public static final int G_POINT = 680;
	//public static final int QUEUE_LEN = 5;
	
	private boolean GOSSIP_OR_NOT = true; //true: use gossip for broadcast
	private boolean UPLOAD_OR_NOT = false; //true: everything will upload to the twitter server
	
	private InetSocketAddress bootaddress;
	private InetAddress bindaddress;
	private int bindport;
	
	protected PastryNodeFactory factory;
	protected Endpoint endpoint;
	protected PastryNode node;
	protected Environment environment;
	protected String userId;
	protected Profile profile;

	//The XML-based database 
	private StatusXMLDB myStatusDB;
	private StatusXMLDB followeeStatusDB;
	private ProfileXMLDB myProfileDB;
	private ProfileXMLDB followeeProfileDB;
	private ProfileXMLDB followerProfileDB;
	private TrafficAnalysisXMLDB trafficAnalysisDB;
	private GossipAnalysisXMLDB gossipAnalysisDB;
	private PartnerXMLDB partnerXMLDB;
	
	private String nodeHandleDir;
	
	private Mailbox mailbox = new Mailbox();
	
	private FixLengthQueue recentFollowers;// = new FixLengthQueue(QUEUE_LEN); 
	
	//used for gossiping
	private PartnerMap partnerMap;
	private BloomFilter bloomFilter;
	
	//private HealthIndex followeeHealthIndex;
	private HealthIndex followerHealthIndex;
	
	private TrafficAnalysisRecorder trafficRecorder;
	
	////////////////////////API with Server//////////////////////////
	TwClientAPI twClientAPI;
	//////////////////////////////////////////////////

	public static String nameExchange(String dirName) {
		return dirName.replaceAll("<", "").replace(">", "").replace(".", "");
	}
	
	public Cuckoo(Environment env, String uId, InetSocketAddress bootaddr, InetAddress bindaddr, int bindport) {
		this.environment = env;
		this.userId = uId;
		this.bootaddress = bootaddr;
		this.bindaddress = bindaddr;
		this.bindport = bindport;

		// Initialize the local XML database
		//System.out.println("dir: " + dir);
		String  dir = "xml/" + userId;
		
		myStatusDB = new StatusXMLDB(dir, dir + "/mystatus.xml", "myStatus");
		followeeStatusDB = new StatusXMLDB(dir, dir + "/followeestatus.xml", "followeeStatus");
		myProfileDB = new ProfileXMLDB(dir, dir + "/myprofile.xml", "myprofile");
		followeeProfileDB = new ProfileXMLDB(dir, dir + "/followeeprofile.xml", "followeeprofile");
		followerProfileDB = new ProfileXMLDB(dir, dir + "/followerprofile.xml", "followerprofile");
		partnerXMLDB = new PartnerXMLDB(dir, dir + "/partner.xml", "partner");
		
		trafficAnalysisDB = new TrafficAnalysisXMLDB(dir+"/trafficanalysis.dat");
		gossipAnalysisDB  = new GossipAnalysisXMLDB(dir,  dir + "/gossipanalysis.xml",  "gossipanalysis");
		
		
		nodeHandleDir = dir + "/followerHandleDir/";
		
		//followeeHealthIndex = new HealthIndex(followeeProfileDB);
		followerHealthIndex = new HealthIndex(followerProfileDB);
		
		trafficRecorder = new TrafficAnalysisRecorder(trafficAnalysisDB, userId);
		
		partnerMap = new PartnerMap(partnerXMLDB);
		bloomFilter = followeeProfileDB.buildBloomFilter();
		
		//load the profile
		profile = myProfileDB.getProfile(userId);
		twClientAPI = new TwClientAPI(profile, trafficRecorder);
		
		gossipOrNot();
		
		try {
			int numOfFollowers = followerProfileDB.getNumOfEntries();
			int qLen = 2 * (((int) Math.ceil(Math.log(numOfFollowers))) + Cuckoo.C);
			recentFollowers = new FixLengthQueue(qLen);
		} catch(Exception e) {
			//do nothing
		}
	}

	//decide whether to use gossip or direct dissemination
	public void gossipOrNot() {
		if(followerProfileDB.getNumOfEntries() >= G_POINT) {
			GOSSIP_OR_NOT = true;
		}
	}

	public void bootstrap() {
		/*
		if(BOOTSTRAP == false) {
			fetchTweetFromServer();
		}*/
		
		try {
			DeterministicNodeIdFactory nidFactory = new DeterministicNodeIdFactory(userId);
			PastryNodeFactory factory = new SocketPastryNodeFactory(nidFactory, bindaddress, bindport, environment);
			node = factory.newNode(nidFactory.generateNodeId());
			
			endpoint = node.buildEndpoint(this, "cuckoo");
			
			///////////////////////////register////////////////////////////////
			endpoint.setDeserializer(new MessageDeserializer() {
			    
			      public Message deserialize(InputBuffer buf, short type, int priority, NodeHandle sender) throws IOException {
			        switch (type) {
			        	case ReqFollowMsg.TYPE:
			        		// just call the special deserialization constructor.
			        		return new ReqFollowMsg(buf, endpoint);
			        	case ReqStatusMsg.TYPE:
			        		return new ReqStatusMsg(buf, endpoint);
			        	case ReqUnfollowMsg.TYPE:
			        		return new ReqUnfollowMsg(buf, endpoint);
			        	case RplFollowErrMsg.TYPE:
			        		return new RplFollowErrMsg(buf, endpoint);
			        	case RplFollowMsg.TYPE:
			        		return new RplFollowMsg(buf, endpoint);
			        	case RplStatusErrMsg.TYPE:
			        		return new RplStatusErrMsg(buf, endpoint);
			        	case RplStatusMsg.TYPE:
			        		return new RplStatusMsg(buf, endpoint);
			        	case DirectUpdateStatusMsg.TYPE:
			        		return new DirectUpdateStatusMsg(buf, endpoint);
			        	case ReqNodeHandleMsg.TYPE:
			        		return new ReqNodeHandleMsg(buf, endpoint);
			        	case RplNodeHandleMsg.TYPE:
			        		return new RplNodeHandleMsg(buf, endpoint);
			        	case RplNodeHandleErrMsg.TYPE:
			        		return new RplNodeHandleErrMsg(buf, endpoint);
			        	case GossipUpdateStatusMsg.TYPE:
			        		return new GossipUpdateStatusMsg(buf, endpoint);
			        	case PartnerCancelMsg.TYPE:
			        		return new PartnerCancelMsg(buf, endpoint);
			        	/*case GossipFeedbackMsg.TYPE:
			        		return new GossipFeedbackMsg(buf, endpoint);*/
			        	case HelloMsg.TYPE:
			        		return new HelloMsg(buf, endpoint);
			        }
			        // maybe throw an IOException if you want
			        return null;
			      }    
			});
			endpoint.register();
			
			///////////////////////////boot////////////////////////////////
			node.boot(bootaddress);
			
			//the node may require sending several messages to fully boot into the ring
			synchronized (node) {
				while (!node.isReady() && !node.joinFailed()) {
					// delay so we don't busy-wait
					node.wait(500);
					// abort if can't join
					if (node.joinFailed()) {
						System.out.println("Could not join the FreePastry ring.  Reason:" + node.joinFailedReason());
					}
				}
			}
			System.out.println("[" + userId + "]:" + Calendar.getInstance().getTime() + " Finished creating new node!");
			
		} catch (IOException ioe) {
			ioe.printStackTrace();
		} catch (InterruptedException ie) {
			ie.printStackTrace();
		}
	}
	
	public void sayHello() {
		LinkedList<String> ll = followeeProfileDB.getInfluentialUserIds();
		Iterator<String> iterator = ll.iterator();
		while(iterator.hasNext()) {
			String feid = iterator.next();
			Id nId = getIdFromString(feid);
			reroute(nId, new HelloMsg(node.getId(), nId, userId, feid, node.getLocalNodeHandle(), Calendar.getInstance().getTimeInMillis(), bloomFilter), null);
		}
	}
	
	public void quit() {
		prepare2quit();
		//endpoint = null;
		node.destroy();
	}
	
	public void die() {
		prepare2quit();
		node.destroy();
		node = null;
		endpoint = null;
	}
	
	public static Id getIdFromString(String uId) {
		DeterministicNodeIdFactory nidFactory = new DeterministicNodeIdFactory(uId);
		return nidFactory.generateNodeId();
	}
	/*
	public void addMyProfile(Profile profile) {
		this.profile = profile;
		myProfileDB.addProfile(profile, null);
	}
	
	public void addMyProfile(String username, String location, String web, String bio) {
		profile = new Profile(userId, username, location, web, bio);
		myProfileDB.addProfile(profile, null);
	}
	*/
	public void follow(String uId) {//uId: userId (the followee)
		//If the user is already followed, return msg;
		if(followeeProfileDB.exist(uId) == true) {
			//timelineFrame.append2FolloweePanel("User " + uId + " is already followed!");
			System.out.println("User " + uId + " is already followed!");
			return;
		}
		
		if(profile == null) {
			System.out.println("ERROR: profile is null!!!");
			System.exit(1);
		}
		
		Id nId = getIdFromString(uId);
		//Pack all the local information into the ReqFollowMsg
		ReqFollowMsg reqFollowMsg = new ReqFollowMsg(node.getId(), nId, userId, uId, node.getLocalNodeHandle(), Calendar.getInstance().getTimeInMillis(), bloomFilter, profile);
		reroute(nId, reqFollowMsg, null);
	}
	
	//this is the main method to call the write
	public void write(String text) {
		endpoint.getEnvironment().getProcessor().process(
				new WriteStatusWorkTask(text),
				new Continuation<Object, Exception>() {
					public void receiveException(Exception arg0) { arg0.printStackTrace(); }
					public void receiveResult(Object arg0) {}
				},
				endpoint.getEnvironment().getSelectorManager(),
				endpoint.getEnvironment().getTimeSource(),
				endpoint.getEnvironment().getLogManager()
		);
		
		/*
		environment.getProcessor().processBlockingIO(
				new WriteStatusWorkTask(
						new Continuation() {
							public void receiveException(Exception arg0) { arg0.printStackTrace(); }
							public void receiveResult(Object arg0) {}
						},
						environment.getSelectorManager(),
						text));
		*/
	}
	
	private void writeStatus(String text) {
		Status s= null;
		if(UPLOAD_OR_NOT == true) {
			//First add on Twitter Server
			String sentText = twClientAPI.uploadStatus(text);
			if(sentText == null) {
				System.out.println("Service reject from Server!!");
			}
		}
		/*
			s = new Status(sentStatus.getStatusId(), text, profile.getUserName(), userId, sentStatus.getTime());
		} else {*/
		s = new Status(String.valueOf(userId+new Date().getTime()), text, profile.getUserName(), userId, new Date().getTime());
		
		myStatusDB.addStatus(s);
		
		disseminate(s);
	}
	
	public void disseminate(Status s) {
		if(DEBUG) {
			System.out.println("userid: " + userId + ", text: " + s.getText());
		}
		//LinkedList<String> followerIds = followerProfileDB.getAllUserIds();
		/*LinkedList<String> followerIds = followerHealthIndex.getHealthyIds();
		
		if(GOSSIP_OR_NOT == true) {
			gossipRandomSelection(followerIds);
		}*/
		LinkedList<String> followerIds = null;
		if(GOSSIP_OR_NOT == false) {
			followerIds = followerProfileDB.getHealthyUserIds();
		} else {
			//followerIds = followerProfileDB.randomSelectHealthyUserIds();
			//System.out.println("Number of random Selection: " + followerIds.size());
			followerIds = new LinkedList<String>();
			followerIds.addAll(recentFollowers.getIds());
			System.out.println("Number of overall Selection: " + followerIds.size());
		}
		
		if(GOSSIP_OR_NOT == true) {
			gossipAnalysisDB.addGossipLog(new GossipLog(new Date().getTime(), s.getStatusId(), 0, 0, 0));
			disseminate(s, followerIds, true, 0, 0);
		} else {
			if(DEBUG) {
				System.out.println("======================================");
				System.out.println("UserId: " + userId);
				System.out.print("Followers: ");
				printStringList(followerIds);
				System.out.println("======================================");
			}
			disseminate(s, followerIds, false, 0, 0);
		}
	}
	
	private void disseminate(Status s, List<String> ids, boolean gossip, long gt, int gh) {
		if(ids == null) {
			return;
		}
		int DoG = (gossip == false) ? 0: 1;
		Iterator<String> iterator = ids.iterator(); 
		while(iterator.hasNext()) {
			String followerId = iterator.next();
			//String path = nodeHandleDir+followerId;
			//NodeHandle nh = (NodeHandle) ObjectSerialize.file2object(path);
			NodeHandle nh = readFile2NodeHandle(followerId);
			
			CuckooMsg cuckoomsg = null;
			if(gossip == false) {
				cuckoomsg = new DirectUpdateStatusMsg(node.getId(), getIdFromString(followerId), userId, followerId, node.getLocalNodeHandle(), Calendar.getInstance().getTimeInMillis(), s);
			} else {
				cuckoomsg = new GossipUpdateStatusMsg(node.getId(), getIdFromString(followerId), userId, followerId, node.getLocalNodeHandle(), Calendar.getInstance().getTimeInMillis(), bloomFilter, s, gt, gh);
			}
			
			if(nh == null) {
				//System.out.println("nodeHandle is null, ask for nodeHandle!!!");

				mailbox.add2mailbox(followerId, cuckoomsg);
				
				Id nId = getIdFromString(followerId);
				
				if(DEBUG) {
					System.out.println("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$");
					System.out.println("$ myUserId = " + userId );
					System.out.println("$ followerId = " + followerId);
					System.out.println("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$");
				}
				
				ReqNodeHandleMsg rnhmsg = new ReqNodeHandleMsg(node.getId(), nId, userId, followerId, node.getLocalNodeHandle(), Calendar.getInstance().getTimeInMillis(), bloomFilter, DoG, s.getAuthorId());
				reroute(nId, rnhmsg, null);
				
			} else { //nh != null => means connect before
				/////////////////Test for exception//////////////////////
				//System.out.println("try to send to node [" + followerId + "]");
				//DirectDeliveryNotification ddn = new DirectDeliveryNotification(mailbox, s, followerId);
				//directRoute(nh, new UpdateStatusMsg(s, node.getId(), getIdFromString(followerId), userId, followerId), ddn);
				//////////////////////////////////////////////////////////
				
				int m = ((PastryNode) node).getLiveness((rice.pastry.NodeHandle)nh);
				if(m == 1) { //alive
					if(DEBUG) {
						System.out.println("m == 1: " + "try to send to node [" + followerId + "]");
					}
					directRoute(nh, cuckoomsg, new DirectDeliveryNotification(mailbox, followerId, this, cuckoomsg), userId, followerId);

				} else { //suspect or dead
					//System.out.println("m != 1: " + "try to send to node [" + followerId + "]");
					deleteNodeHandle(followerId);
					
					mailbox.add2mailbox(followerId, cuckoomsg);
					
					//mailbox.printMailbox();
					//System.out.println("-------" + nh.getId().toString());
					Id nId = getIdFromString(followerId);
					//System.out.println("-----------followerId: " + followerId);
					//System.out.println("------------------nId: " + nId.toString());
					//System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
					ReqNodeHandleMsg rnhmsg = new ReqNodeHandleMsg(node.getId(), nId, userId, followerId, node.getLocalNodeHandle(), Calendar.getInstance().getTimeInMillis(), bloomFilter, DoG, s.getAuthorId());
					
					reroute(nId, rnhmsg, null);
				}
			}
		}
		followerHealthIndex.writeBack();
	}
	
	/*
	private void sendGossipAck(String oid, GossipFeedbackMsg gfm) {
		NodeHandle nh = readFile2NodeHandle(oid);
		Id nId = getIdFromString(oid);
		
		if(nh == null) {
			mailbox.add2mailbox(oid, gfm);
			ReqNodeHandleMsg rnhmsg = new ReqNodeHandleMsg(node.getId(), nId, userId, oid, node.getLocalNodeHandle(), Calendar.getInstance().getTimeInMillis(), bloomFilter, 2, userId);
			reroute(nId, rnhmsg, null);
		} else { //nh != null => means connect before
			int m = ((PastryNode) node).getLiveness((rice.pastry.NodeHandle)nh);
			if(m == 1) { //alive
				directRoute(nh, gfm, new DirectDeliveryNotification(mailbox, oid, this, gfm), userId, oid);
			} else { //suspect or dead
				mailbox.add2mailbox(oid, gfm);
				ReqNodeHandleMsg rnhmsg = new ReqNodeHandleMsg(node.getId(), nId, userId, oid, node.getLocalNodeHandle(), Calendar.getInstance().getTimeInMillis(), bloomFilter, 2, userId);
				reroute(nId, rnhmsg, null);
			}
		}
	}
	*/
	
	public void gossipRandomSelection(LinkedList<String> followerIds) {
		int gNum = (int) (Math.log( followerProfileDB.getNumOfEntries() ) + C);
		//int gNum = (int) (Math.log( followerHealthIndex.getNum() ) + C);
		gossipRandomSelection(followerIds, gNum);
	}
	
	private void gossipRandomSelection(LinkedList<String> followerIds, int num) {
		if(followerIds.size() <= num) {
			return;
		} else {
			int l = followerIds.size()-num;
			for(int i=0; i<l; i++) {
				int ran = (int) (Math.random()*(followerIds.size()));
				followerIds.remove(ran);
			}
		}
	}

	//For followees!
	public void addFollower(Profile fprofile) {
		followerProfileDB.addProfile(fprofile, null);
	}
	
	public void updateFolloweeStatus() {
		LinkedList<UTPair> followees = followeeProfileDB.getAllUTPair();
		//If no followee
		if(followees == null || followees.size() == 0) {
			System.out.println("no followee!!");
			return;
		}
		
		//DEBUG
		//UTPair.printUTPairList(followees);
		
		UTPair pair;
		Iterator<UTPair> iterator = followees.iterator(); 
		while(iterator.hasNext()) {
			pair = iterator.next();
			updateFolloweeStatus(pair);
		}		
	}
	
	public void updateFolloweeStatus(String userId) {
		UTPair pair = followeeProfileDB.getUTPair(userId);
		if(pair != null) {
			updateFolloweeStatus(pair);
		}
	}
	
	private void updateFolloweeStatus(UTPair pair) { //userId and last time
		//pair.print();
		Id nId = getIdFromString(pair.userId);
		ReqStatusMsg reqStatusMsg = new ReqStatusMsg(node.getId(), nId, userId, pair.userId, node.getLocalNodeHandle(), Calendar.getInstance().getTimeInMillis(), bloomFilter, pair.time);
		
		reroute(nId, reqStatusMsg, null);	
		//statusPendings.add(pair.userId);
	}
	
	/**
	 * @Override: required by the "Application" interface
	 * Called when we receive a message.
	 */
	public void deliver(Id id, Message message) {
		endpoint.getEnvironment().getProcessor().process(
				/*new DeliverWorkRequest(
						new Continuation() {
							public void receiveException(Exception arg0) { arg0.printStackTrace(); }
							public void receiveResult(Object arg0) {}
						},
						environment.getSelectorManager(),
						message,
						this));*/
				new DeliverWorkTask(message, this),
				new Continuation<Object, Exception>() {
					public void receiveException(Exception arg0) { arg0.printStackTrace(); }
					public void receiveResult(Object arg0) {}
				},
				endpoint.getEnvironment().getSelectorManager(),
				endpoint.getEnvironment().getTimeSource(),
				endpoint.getEnvironment().getLogManager()
		);
	}

	/**
	 * Called when a message is traveling along your path. 
	 */
	public boolean forward(RouteMessage message) {
		try {
			Message mesg = message.getMessage(endpoint.getDeserializer());
			
			/*if(mesg instanceof CuckooMsg) {
				((CuckooMsg) mesg).addHops();
				message.setMessage(mesg);
			}*/
			endpoint.getEnvironment().getProcessor().process(
					new ForwardWorkTask(mesg, message),
					new Continuation<Object, Exception>() {
						public void receiveException(Exception arg0) { arg0.printStackTrace(); }
						public void receiveResult(Object arg0) {}
					},
					endpoint.getEnvironment().getSelectorManager(),
					endpoint.getEnvironment().getTimeSource(),
					endpoint.getEnvironment().getLogManager()
			);
			/*environment.getProcessor().processBlockingIO(
				new ForwardWorkRequest(
						new Continuation() {
							public void receiveException(Exception arg0) { arg0.printStackTrace(); }
							public void receiveResult(Object arg0) {}
						},
						environment.getSelectorManager(),
						mesg,
						message));*/
		} catch (IOException e) {
			e.printStackTrace();
		}
		return true;
	}

	/**
	 * Called when you hear about a new neighbor. Don't worry about this method
	 * for now.
	 */
	public void update(NodeHandle arg0, boolean arg1) {
		
	}

	/**
	 * Called to directly send a message to the nh
	 * The receiver receive the msg by "deliver" function
	 */
	private void directRoute(NodeHandle nh, Message msg, DeliveryNotification dn, String fid, String tid) {
		//System.out.println(this+" sending direct to "+nh);
		try {
			if(dn != null) {
				endpoint.route(null, msg, nh, dn);
			} else {
				endpoint.route(null, msg, nh);
			}
			//---traffic log
			trafficRecorder.record(msg, Calendar.getInstance().getTimeInMillis(), "p", "o", fid, tid, "n");
		} catch(NullPointerException npe) {
			npe.printStackTrace();
			System.out.println("msg: " + msg);
			System.out.println("nh: " + nh);
			System.out.println("dn: " + dn);
		}
	}
	
	public void reroute(Id destId, CuckooMsg msg, DeliveryNotification dn) {
		endpoint.route(destId, msg, null, dn);
		trafficRecorder.record(msg, Calendar.getInstance().getTimeInMillis(), "p", "o", msg.getFromUserId(), msg.getToUserId(), "n");
	}
	
	//return the NodeHandle of a specific user with uid
	public NodeHandle readFile2NodeHandle(String uid) {
		return (NodeHandle) ObjectSerialize.file2object(nodeHandleDir+uid);
	}
	
	//write NodeHandle to file (serialization)
	public void writeNodeHandle2file(NodeHandle nh, String uid) {
		ObjectSerialize.object2file(nh, nodeHandleDir+uid);
	}
	
	public void deleteNodeHandle(String id) {
		File file = new File(nodeHandleDir+id);
		if(file.exists()) {
			file.delete();
		} 
		//maybe delete by sendFail
		  /*else {     
			System.out.println("ERROR: the nodeHandle file does not exist!!");     
		} */    
	}
		
	public String toString() {
		return "Cuckoo (nodeId: " + endpoint.getId() + ", userId: " + userId + ")";
	}
	
	public void prepare2quit() {
		//System.out.println("[" + userId + "]: Bye, Bye!!");
		//write back the healthIndex
		followerHealthIndex.writeBack();
		recentFollowers.clearQueue();
		mailbox.clearMailBox();
		//followeeHealthIndex.writeBack();
		//System.exit(1);
	}
	
	public void monitorPassingMsg(CuckooPiggybackMsg cpmsg) {
		monitorPassingGuy(cpmsg.getFromUserId(), cpmsg.getBloomFilter(), cpmsg.getNodeHandle());
	}
	
	private void monitorPassingGuy(String pid, BloomFilter bf, NodeHandle nh) {
		LinkedList<String> followeeIds = followeeProfileDB.getAllUserIds();
		Iterator<String> iterator = followeeIds.iterator();
		LinkedList<String> ll = new LinkedList<String>();
		while(iterator.hasNext()) {
			String fid = iterator.next();
			if(bf.check(fid.getBytes()) == true) {
				//System.out.println("common interest: " + fid);
				/*if(p == null) {
					p = new Partner(fid, pid, fid, 1, HealthIndex.INIT_BLOOD);
				} else {
					//p.addCommonId(fid);
					p.addSimilarity();
				}*/
				ll.add(fid);
			}
		}
		iterator = ll.iterator();
		while(iterator.hasNext()) {
			partnerMap.monitorPartner(iterator.next(), pid);
		}
	}
	
	/*
	public String generatePartners() {
		LinkedList<String> followerIds = followerHealthIndex.getHealthyIds();
		gossipRandomSelection(followerIds, PartnerMap.NUM_OF_PARTNERS);
		
		String ids = "";
		Iterator<String> iterator = followerIds.iterator();
		while(iterator.hasNext()) {
			ids = ids + iterator.next() + "#";
		}
		//System.out.println("generatePartners ------------------ " + ids);
		return ids;
	}*/
	/*
	public void initPartnerMap(String fid, String pids) {
		String[] split = pids.split("#");
		for(int i=0; i<split.length; i++) {
			partnerMap.initAdd(fid, new Partner(fid, split[i], fid, 1, HealthIndex.INCR_BLOOD));		
		}
		partnerMap.writeBack();
	}
	*/
	public void increaseGossipSent(String stId) {
		gossipAnalysisDB.increaseSentMsg(stId);
	}
	
	public void increaseGossipRecv(String stId) {
		gossipAnalysisDB.increaseRecvMsg(stId);
	}
	
	public void increaseFollowerHealth(String fid) {
		followerHealthIndex.increaseHealthIndex(fid);
	}
	
	public void decreaseFollowerHealth(String fid) {
		followerHealthIndex.decreaseHealthIndex(fid);
	}
	
	public void increasePartnerHealth(String fid, String pid) {
		partnerMap.increaseHealth(fid, pid);
	}
	
	public void decreasePartnerHealth(String fid, String pid) {
		partnerMap.decreaseHealth(fid, pid);
	}
	
	public void fetchTweetFromServer() {
		//new Thread() {
		//	public void run() {

		Status[] statuses = null;
		statuses = twClientAPI.requestRecentUpdates(followeeStatusDB.getLastTime());
		
		//---traffic log
		//trafficRecorder.record(new ServerRequestMsg(node.getId(), userId, ""), Calendar.getInstance().getTimeInMillis(), "s", "o", userId, "server", "n");
		//trafficRecorder.record(new ServerGetStatusesMsg(stl, node.getId(), userId), Calendar.getInstance().getTimeInMillis(), "s", "i", "server", userId, "n");		
		
		if(statuses != null) {
			for(int i=0; i<statuses.length; i++) {
				if(userId.compareTo(String.valueOf(statuses[i].getAuthorId())) != 0) {
					followeeStatusDB.addStatus(statuses[i]);
					followeeProfileDB.updateTime(statuses[i].getAuthorId(), statuses[i].getTime());
				}
			}
		}
		
			//}
		//}.start();
	}
	
	/*
	public void fetchTweetFromServer(String userid) {
		Status[] statuses = null;
		statuses = twClientAPI.requestRecentUpdates(followeeProfileDB.getLastTime(userid));
		
		if(statuses == null) {return;}
		for(int i=0; i<statuses.length; i++) {
			if(userId.compareTo(String.valueOf(statuses[i].getAuthorId())) != 0) {
				followeeStatusDB.addStatus(statuses[i]);
				followeeProfileDB.updateTime(statuses[i].getAuthorId(), statuses[i].getTime());
			}
		}
	}
	*/
	//////////////Getters/////////////////
	public Node getNode() {
		return node;
	}
	
	public String getUserId() {
		return userId;
	}
	
	public Id getNodeId() {
		return node.getId();
	}
	
	public NodeHandle getNodeHandle() {
		return node.getLocalNodeHandle();
	}
	
	public TrafficAnalysisXMLDB getTrafficAnalysisXMLDB() {
		return trafficAnalysisDB;
	}
	
	public GossipAnalysisXMLDB getGossipAnalysisXMLDB() {
		return gossipAnalysisDB;
	}
	/*
	public Status[] getFolloweeStatus() {
		return followeeStatusDB.getRecentStatus(BULK);
	}*/
	/*
	public LinkedList<Profile> getFolloweeProfile() {
		return followeeProfileDB.getAllProfile();
	}
	
	public LinkedList<Profile> getFollowerProfile() {
		return followerProfileDB.getAllProfile();
	}*/
	
	public Status[] getMyStatus() {
		Status[] ss = myStatusDB.getRecentStatuses(BULK);
		if(ss == null || ss.length == 0) {
			return null;
		}
		return ss;
	}
	
	public BloomFilter getBloomFilter() {
		return bloomFilter;
	}
	
	public boolean getGossipOrNot() {
		return GOSSIP_OR_NOT;
	}
	
	public boolean getUploadOrNot() {
		return UPLOAD_OR_NOT;
	}
	
	public Profile getProfile() {
		return profile;
	}
	
	//////////////Printers/////////////////
	
	public void printProfile() {
		myProfileDB.xmlParsePrint();
	}
	
	public void printMyStatus() {
		StatusXMLDB.printStatuses(myStatusDB.getRecentStatuses(BULK));
	}
	
	public void printFolloweeStatus() {
		StatusXMLDB.printStatuses(followeeStatusDB.getRecentStatuses(BULK));
	}
	
	public void printFollowers() {
		followerProfileDB.xmlParsePrint();
	}
	
	public void printFollowees() {
		followeeProfileDB.xmlParsePrint();
	}
	
	public void printGossipOrNot() {
		System.out.println(GOSSIP_OR_NOT);
	}
	
	public static void printStringList(List<String> l) {
		Iterator<String> iterator = l.iterator();
		while(iterator.hasNext()) {
			System.out.print(iterator.next() + "  ");
		}
		System.out.println();
	}
	/*
	public void printFolloweeHealthIndex() {
		System.out.println("----------Followee Health Index----------");
		followeeHealthIndex.printHeathIndice();
	}
	*/
	public void printFollowerHealthIndex() {
		System.out.println("----------Follower Health Index----------");
		followerHealthIndex.printHeathIndice();
	}
	/*
	public void printPartnerMap() {
		System.out.println("---------------Partner Map---------------");
		partnerMap.printPartnerMap();
	}
	*/
	////////Setters///////////////////////
	//public void setBootstrap(boolean isBootstrap) {
	//	BOOTSTRAP = isBootstrap;
	//}
	
	public void setGossipOrNot(boolean g) {
		GOSSIP_OR_NOT = g;
	}
	
	public void setUploadOrNot(boolean u) {
		UPLOAD_OR_NOT = u;
	}
	
	public boolean isFollower(String fid) {
		return followerProfileDB.exist(fid);
	}
	
	//the core of the runnable interface
	public void run() {
		Tweet tweet = new Tweet();
		tweet.start();
	}
	
	/*********************** Inner classes **************************/
	
	public class DeliverWorkTask implements Executable<Object, Exception> {
		private Message message;
		private Cuckoo cuckoo;
		
		public DeliverWorkTask(Message m, Cuckoo cuc) {
			super();
			message = m;
			cuckoo = cuc;
		}
		
		/*
		public DeliverWorkTask(Continuation<Object,java.lang.Exception> c, SelectorManager sm, Message m, Cuckoo cuc) {
			super(c, sm);
			message = m;
			cuckoo = cuc;
		}
		*/
		
		public Object execute() throws Exception {
			if(!(message instanceof CuckooMsg)) {
				return null;
			}
			
			CuckooMsg cm = (CuckooMsg) message;
			if( cm.getFromUserId() == userId ) {
				/*
				if(DEBUG) {
					System.out.println("sent by myself..., message type: " + cm.getType());
					if( cm instanceof ReqNodeHandleMsg ) {
						System.out.println("RequestNodeHandleMsg: toUserId -> " + cm.getToUserId() + ", myUserId ->" + userId);
					}
				}
				*/
				return null;
			}
			//Get the nodeHandler from the msg
			NodeHandle nh = cm.getNodeHandle();
			/*if(! (cm instanceof GossipFeedbackMsg)) {
				trafficRecorder.record(cm, Calendar.getInstance().getTimeInMillis(), "p", "i", cm.getFromUserId(), cm.getToUserId(), "n");
			}
			*/	
			if (message instanceof ReqFollowMsg) {
				//System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive ReqFollowMsg!");
				ReqFollowMsg msg = (ReqFollowMsg) message;

				if(msg.getToUserId().compareTo(userId) != 0) {
					System.out.println("The target of ReqFollowMsg is not online, my userId: " + userId + ", msg.userId: " + msg.getToUserId() + "***");
					
					//Return back the error msg to tell the offline condition!
					RplFollowErrMsg rplFollowErrMsg = new RplFollowErrMsg(node.getId(), msg.getFrom(), userId, msg.getFromUserId(), node.getLocalNodeHandle(), Calendar.getInstance().getTimeInMillis(), bloomFilter, msg.getToUserId());
					directRoute(nh, rplFollowErrMsg, null, rplFollowErrMsg.getFromUserId(), rplFollowErrMsg.getToUserId());
					return null;
				}
				
				//Create Reply-Follow-Message that contains the newest profile; 
				RplFollowMsg profilemsg = new RplFollowMsg(node.getId(), msg.getFrom(), userId, msg.getFromUserId(), node.getLocalNodeHandle(), Calendar.getInstance().getTimeInMillis(), profile, ""/*generatePartners()*/); //here just because this will never be used in the emulation
				addFollower(msg.getProfile()); //here init the BLOOD automatically
				writeNodeHandle2file(nh, msg.getProfile().getUserId());
				followerHealthIndex.increaseHealthIndex(msg.getFromUserId());
				
				msg.getProfile().print();
				
				//Send the Reply-Follow-Message to the follower as ACK;
				directRoute(nh, profilemsg, null, profilemsg.getFromUserId(), profilemsg.getToUserId());
				
			} else if(message instanceof RplFollowMsg) {
				//follow on Twitter: seems no need in the experiments (emulation)
				
				//System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive RplFollowMsg!");
				/*RplFollowMsg msg = (RplFollowMsg) message;
				
				if(msg.getToUserId().compareTo(userId) != 0) {
					System.out.println("The target of RqlFollowMsg (" + msg.getToUserId() + ") is not online.");
					return null; //do nothing!!
				}
				
				followeeProfileDB.addProfile(msg.getProfile()); //here init the BLOOD automatically
				bloomFilter.add(msg.getProfile().getUserId().getBytes());
				//followeeHealthIndex.addHealthIndex(msg.getProfile().getUserId());
				initPartnerMap(msg.getProfile().getUserId(), msg.getPartnerIds());
				
				
				
				try {
					Twitter.User tuser = twitter.getUser(Long.valueOf(msg.getFromUserId()));
					//---traffic log
					trafficRecorder.record(new ServerRequestMsg(node.getId(), userId, msg.getFromUserId()), Calendar.getInstance().getTimeInMillis(), "s", "o", userId, "server", "n");
					twitter.follow(tuser.screenName);
					//---traffic log
					trafficRecorder.record(new ServerRequestMsg(node.getId(), userId, tuser.screenName), Calendar.getInstance().getTimeInMillis(), "s", "o", userId, "server", "n");
					
				} catch(Exception e) {
					handleTwitterException(e);
				}*/
				
			} else if(message instanceof ReqStatusMsg) {
				//System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive ReqStatusMsg!");
				ReqStatusMsg msg = (ReqStatusMsg) message;
				
				if(msg.getToUserId().compareTo(userId) != 0) {
					System.out.println("The target of ReqStatusMsg is offline, my userId: " + userId + ", msg.userId: " + msg.getToUserId() + "******");
					
					//Return back the msg that tells the offline condition!
					RplStatusErrMsg rplStatusErrMsg = new RplStatusErrMsg(node.getId(), msg.getFrom(), userId, msg.getFromUserId(), node.getLocalNodeHandle(), Calendar.getInstance().getTimeInMillis(), bloomFilter, msg.getToUserId(), msg.getAcHops(), msg.getAcTime(), msg.getHops(), (Calendar.getInstance().getTimeInMillis()-msg.getSTime()));
					directRoute(nh, rplStatusErrMsg, null, rplStatusErrMsg.getFromUserId(), rplStatusErrMsg.getToUserId());
					return null;
				}
				
				writeNodeHandle2file(nh, msg.getFromUserId());
				followerHealthIndex.increaseHealthIndex(msg.getFromUserId());
				
				Status[] status = myStatusDB.getStatuses(msg.getTime());
				
				if(DEBUG) {
					System.out.println("msg.getFrom() " + node.getId());
					System.out.println("msg.getFromUserId() " + userId);
					System.out.println("msg.getFrom() " + msg.getFrom());
					System.out.println("msg.getFromUserId() " + msg.getFromUserId());
					System.out.println("msg.getTo() " + msg.getTo());
					System.out.println("msg.getToUserId() " + msg.getToUserId());
				}
				
				RplStatusMsg rplStatusMsg = new RplStatusMsg(node.getId(), msg.getFrom(), userId, msg.getFromUserId(), node.getLocalNodeHandle(), Calendar.getInstance().getTimeInMillis(), userId, status, msg.getAcHops(), msg.getAcTime(), msg.getHops(), (Calendar.getInstance().getTimeInMillis()-msg.getSTime()));
				
				//Send back the microblogs -- not care about whether that one can receive
				directRoute(nh, rplStatusMsg, null, rplStatusMsg.getFromUserId(), rplStatusMsg.getToUserId());
				
			} else if(message instanceof RplStatusMsg) {
				//System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive RplStatusMsg!");
				RplStatusMsg msg = (RplStatusMsg) message;
				
				String authorId = msg.getAuthorId();
				if(followeeProfileDB.exist(authorId) == false) {
					System.out.println("error!!! I don't need the msg!!! (in RplStatusMsg)");
				}
				//The microblogs are in right track
				String authorName = followeeProfileDB.getNameFromId(authorId);
				
				String[] stid = msg.getRecentStids();
				String[] text = msg.getRecentTexts();
				long[] time = msg.getRecentTimes();
				
				if(stid == null) { return null; }
				
				Status ns = null;
				for(int i=stid.length-1; i>=0; i--) {
					ns = new Status(stid[i], text[i], authorName, authorId, time[i]);
					followeeStatusDB.addStatus(ns);
				}
				followeeProfileDB.updateTime(ns.getAuthorId(), String.valueOf(ns.getTime()));
				//followeeHealthIndex.healthIncrease(authorId);
				
			} else if(message instanceof DirectUpdateStatusMsg) {
				//System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive DirectUpdateStatusMsg!");
				DirectUpdateStatusMsg msg = (DirectUpdateStatusMsg) message;
				
				//System.out.println("DirectUpdateStatusMsg.getFromUserId: --" + msg.getFromUserId());
				//System.out.println("DirectUpdateStatusMsg.getFrom: " + msg.getFrom().toString());
				//System.out.println("DirectUpdateStatusMsg.getToUserId: ----" + msg.getToUserId());
				//System.out.println("DirectUpdateStatusMsg.getTo: " + msg.getTo());
				//System.out.println("DirectUpdateStatusMsg.text: -----------" + msg.getText());
				
				String authorId = msg.getAuthorId();
				
				if(followeeProfileDB.exist(authorId) == false) {
					System.out.println("***********************************************************");
					System.out.println("| error!!! I don't need the msg!!! (in DirectUpdateStatusMsg)");
					System.out.println("| DirectUpdateStatusMsg.getFromUserId: --" + msg.getFromUserId());
					System.out.println("| DirectUpdateStatusMsg.getToUserId: ----" + msg.getToUserId());
					System.out.println("| DirectUpdateStatusMsg.text: -----------" + msg.getText());
					System.out.println("| DirectUpdateStatusMsg.author: ---------" + msg.getAuthorId());
					System.out.println("***********************************************************");
					ReqUnfollowMsg reqUnfollowMsg = new ReqUnfollowMsg(node.getId(), msg.getFrom(), userId, msg.getFromUserId(), node.getLocalNodeHandle(), Calendar.getInstance().getTimeInMillis(), bloomFilter);
					reroute(msg.getFrom(), reqUnfollowMsg, null);		
					return null;
				}
				
				String authorName = followeeProfileDB.getNameFromId(authorId);
				//debugging
				//System.out.println(userId + " receive " + msg.getText() + " from " + authorName);
				
				Status ns = new Status(msg.getStatusId(), msg.getText(), authorName, authorId, msg.getTime());
				followeeStatusDB.addStatus(ns);
				followeeProfileDB.updateTime(ns.getAuthorId(), String.valueOf(ns.getTime()));
				//followeeHealthIndex.healthIncrease(authorId);
				
			} else if(message instanceof GossipUpdateStatusMsg) {
				//System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive GossipUpdateStatusMsg!");
				GossipUpdateStatusMsg msg = (GossipUpdateStatusMsg) message;
				
				//System.out.println("GossipUpdateStatusMsg.getFromUserId: --" + msg.getFromUserId());
				//System.out.println("GossipUpdateStatusMsg.getToUserId: ----" + msg.getToUserId());
				//System.out.println("GossipUpdateStatusMsg.text: -----------" + msg.getText());
				
				String authorId = msg.getAuthorId();
				
				if(followeeProfileDB.exist(authorId) == false) {
					System.out.println("error!!! I don't need the msg!!! (in GossipUpdateStatusMsg)");
					//cancel the partnership
					PartnerCancelMsg partnerCancelMsg = new PartnerCancelMsg(node.getId(), msg.getFrom(), userId, msg.getFromUserId(), node.getLocalNodeHandle(), Calendar.getInstance().getTimeInMillis(), authorId);
					directRoute(nh, partnerCancelMsg, null, userId, msg.getFromUserId());
					return null;
				}
				
				String authorName = followeeProfileDB.getNameFromId(authorId);
				Status ns = new Status(msg.getStatusId(), msg.getText(), authorName, authorId, msg.getTime());
				if(followeeStatusDB.recentExist(ns) == false) {
					
					//System.out.println("------------------------------------");
					//System.out.println("| now hops = " + ((CuckooMsg) mesg).getHops() + "|");
					//System.out.println("------------------------------------");
					
					/*first show on the interface and add to local db*/
					followeeStatusDB.addStatus(ns);
					followeeProfileDB.updateTime(ns.getAuthorId(), String.valueOf(ns.getTime()));
					//followeeHealthIndex.healthIncrease(authorId);
					
					/*gossip now, ahahahahaha!*/
					List<String> partners = partnerMap.getPartners(msg.getAuthorId());
					
					//System.out.println("///////////////////////////////////////////////////////////");
					//printStringList(partners);
					long gt = msg.getGossipTime() + (Calendar.getInstance().getTimeInMillis() - msg.getSTime());
					int gh = msg.getGossipHops() + 1;
					
					if(DEBUG) {
						System.out.println("------------------------------------");
						System.out.println("| gossip time = " + msg.getGossipTime() + "|");
						System.out.println("| gossip hops = " + msg.getGossipHops() + "|");
						System.out.println("------------------------------------");
					}
					
					disseminate(ns, partners, true, gt, gh);
					
					/*send back the recv msg to the original user :-(*/
					//sendGossipAck(authorId, new GossipFeedbackMsg(node.getId(), getIdFromString(authorId), userId, authorId, node.getLocalNodeHandle(), Calendar.getInstance().getTimeInMillis(), msg.getStatusId()));
					
					partnerMap.writeBack(PartnerMap.FOR_SEND);
					
				} else {
					if(DEBUG) {
						System.out.println("node [" + userId + "]: duplicate status: " + ns.getText());
					}
					//System.out.println("Redundant message: from [" + msg.getFromUserId() + "] to [" + msg.getToUserId() + "], text: " + msg.getText());
					trafficRecorder.record(cm, Calendar.getInstance().getTimeInMillis(), "p", "i", cm.getFromUserId(), cm.getToUserId(), "r");
				}
			} else if(message instanceof RplFollowErrMsg) {
				//System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive RplFollowErrMsg!");
				RplFollowErrMsg msg = (RplFollowErrMsg) message;
				
				if(msg.getToUserId().compareTo(userId) != 0) {
					System.out.println("The target of RqlFollowErrMsg (" + msg.getToUserId() + ") is not online.");
					return null;
				}
				/*
				//The follow target is offline, just request to server
				Twitter.User tuser = null;
				try {
					tuser = twitter.getUser(Long.valueOf(msg.getUserId()));
					//---traffic log
					trafficRecorder.record(new ServerRequestMsg(node.getId(), userId, msg.getUserId()), Calendar.getInstance().getTimeInMillis(), "s", "o", userId, "server", "n");
					twitter.follow(tuser.screenName);
					//---traffic log
					trafficRecorder.record(new ServerRequestMsg(node.getId(), userId, tuser.screenName), Calendar.getInstance().getTimeInMillis(), "s", "o", userId, "server", "n");
				} catch(Exception e) {
					handleTwitterException(e);
				}
				Profile p = new Profile(tuser);
				
				followeeProfileDB.addProfile(p);
				followeeHealthIndex.addHealthIndex(p.getUserId());
				followeeHealthIndex.healthDecrease(p.getUserId());
				if(timelineFrame != null) {timelineFrame.append2FolloweePanel(p.toString());}
				*/
			} else if(message instanceof RplStatusErrMsg) {
				//System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive RplStatusErrMsg!");
				//The request target is offline, re-evaluate the H-index
				RplStatusErrMsg msg = (RplStatusErrMsg) message;
				
				if(msg.getToUserId().compareTo(userId) == 0) {
					//followeeHealthIndex.healthDecrease(msg.getUserId());
					//fetchTweetFromServer(msg.getUserId());
				}
			} else if(message instanceof ReqUnfollowMsg) {
				/*
				 System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive ReqUnfollowMsg from [" + ((ReqUnfollowMsg) message).getFromUserId() + "]!!");
				ReqUnfollowMsg msg = (ReqUnfollowMsg) message;
				
				if(msg.getToUserId().compareTo(userId) != 0) {
					return null; //do nothing, just discard.
				}
				followerProfileDB.deleteProfile(msg.getFromUserId());
				followerHealthIndex.deleteHealthIndex(msg.getFromUserId());
				*/
			} else if(message instanceof ReqNodeHandleMsg) {
				//System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive ReqNodeHandleMsg!");
				ReqNodeHandleMsg msg = (ReqNodeHandleMsg) message;
				
				//System.out.println("ReqNodeHandleMsg.getFromUserId: " + msg.getFromUserId());
				//System.out.println("ReqNodeHandleMsg.getFrom: " + msg.getFrom().toString());
				//System.out.println("ReqNodeHandleMsg.getToUserId: " + msg.getToUserId());
				//System.out.println("ReqNodeHandleMsg.getTo: " + msg.getTo());
				
				if(msg.getToUserId().compareTo(userId) != 0) {
					//System.out.println("target offline!");
					//System.out.println("msg.getToUserId()" + msg.getToUserId() +", userId: " + userId);
					
					directRoute(nh, new RplNodeHandleErrMsg(node.getId(), msg.getFrom(), userId, msg.getFromUserId(), node.getLocalNodeHandle(), Calendar.getInstance().getTimeInMillis(), bloomFilter, msg.getToUserId(), msg.getDoG(), msg.getAuthorId()), null, userId, msg.getFromUserId());
					return null;
				}
				directRoute(nh, new RplNodeHandleMsg(node.getId(), msg.getFrom(), userId, msg.getFromUserId(), node.getLocalNodeHandle(), Calendar.getInstance().getTimeInMillis()), null, userId, msg.getFromUserId());
				
			} else if(message instanceof RplNodeHandleMsg) {
				//System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive RplNodeHandleMsg!");
				RplNodeHandleMsg msg = (RplNodeHandleMsg) message;
				
				//System.out.println("RplNodeHandleMsg.getFromUserId: " + msg.getFromUserId());
				//System.out.println("RplNodeHandleMsg.getFrom: " + msg.getFrom().toString());
				//System.out.println("RplNodeHandleMsg.getToUserId: " + msg.getToUserId());
				//System.out.println("RplNodeHandleMsg.getTo: " + msg.getTo());
				
				if(msg.getToUserId().compareTo(userId) == 0) {
					writeNodeHandle2file(nh, msg.getFromUserId());
					
					//clear the log send
					LinkedList<CuckooMsg> alist = mailbox.getQueuedCuckooMsgs(msg.getFromUserId());
					if(alist != null) {
						while(!alist.isEmpty()) {
							CuckooMsg cmsg = alist.removeFirst();
							cmsg.setTime(Calendar.getInstance().getTimeInMillis());
							directRoute(nh, cmsg, new DirectDeliveryNotification(mailbox, msg.getFromUserId(), cuckoo, cmsg), userId, msg.getFromUserId());
						}
					} else {
						System.out.println(msg.getFromUserId() + "in mailbox is null");
					}
				}
			} else if(message instanceof HelloMsg) {
				HelloMsg msg = (HelloMsg) message;
				if(msg.getToUserId().compareTo(userId) == 0) {
					writeNodeHandle2file(nh, msg.getFromUserId());
					recentFollowers.add2queue(msg.getFromUserId());
				} //else do nothing
				
			} else if(message instanceof RplNodeHandleErrMsg) {
				//System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive RplNodeHandleErrMsg!");
				RplNodeHandleErrMsg msg = (RplNodeHandleErrMsg) message;
				
				//System.out.println("RplNodeHandleErrMsg.getFromUserId: " + msg.getFromUserId());
				//System.out.println("RplNodeHandleErrMsg.getFrom: " + msg.getFrom().toString());
				//System.out.println("RplNodeHandleErrMsg.getToUserId: " + msg.getToUserId());
				//System.out.println("RplNodeHandleErrMsg.getTo: " + msg.getTo());
				if(msg.getToUserId().compareTo(userId) == 0) {
					int dog = msg.getDoG();
					String oid = msg.getOriginId();
					String aid = msg.getAuthorId();
					if(dog == 0) { // direct dissemination fail
						decreaseFollowerHealth(oid);
					} else if(dog == 1) { //gossip dissemination fail
						if(oid.compareTo(aid) == 0) {
							decreaseFollowerHealth(oid);
						} else {
							decreasePartnerHealth(aid, oid);
						}
					} else if(dog == 2) { //ack dissemination
						//do nothing
						System.out.println("We do nothing for ack here");
					}
				}
			} else if(message instanceof PartnerCancelMsg) {
				//need not here
				
				//System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive PartnerCancelMsg!");
				PartnerCancelMsg msg = (PartnerCancelMsg) message;
				
				if(msg.getToUserId().compareTo(userId) == 0) {
					//partnerMap.deletePartner(msg.getCommonFolloweeId(), msg.getFromUserId());
					System.out.println("should be partner cancel here!!!");
				}
			} /*else if(message instanceof GossipFeedbackMsg) {
				//System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive GossipFeedbackMsg!");
				GossipFeedbackMsg msg = (GossipFeedbackMsg) message;
				
				if(msg.getToUserId().compareTo(userId) == 0) { // else do nothing! (just discard)
					increaseGossipRecv(msg.getStatusId());
				}
			}*/
			return null;
		}
	}
	
	public class ForwardWorkTask implements Executable<Object, Exception> {
		private Message mesg;
		private RouteMessage message;
		
		public ForwardWorkTask(Message m, RouteMessage rm) {
			super();
			mesg = m;
			message = rm;
		}
		
		/*
		public ForwardWorkRequest(Continuation<Object,java.lang.Exception> c, SelectorManager sm, Message m, RouteMessage rm) {
			super(c, sm);
			mesg = m;
			message = rm;
		}
		*/
		
		public Object execute() throws Exception {
			if(mesg instanceof CuckooMsg) {
				((CuckooMsg) mesg).addHops();
				message.setMessage(mesg);
				
				if(DEBUG) {
				if(((CuckooMsg) mesg).getToUserId().compareTo(userId)!=0 && ((CuckooMsg) mesg).getFromUserId().compareTo(userId)!=0) {
					System.out.println("------------------------------------");
					System.out.println("get mesg type: " + ((CuckooMsg) mesg).getType());
					System.out.println("| myid " + userId);
					System.out.println("| from " + ((CuckooMsg) mesg).getFromUserId());
					System.out.println("| to   " + ((CuckooMsg) mesg).getToUserId());
					System.out.println("| now hops = " + ((CuckooMsg) mesg).getHops() + "|");
					System.out.println("------------------------------------");
				}
				}
			}
			
			if(mesg instanceof ReqStatusMsg) {
				ReqStatusMsg rsm = (ReqStatusMsg) mesg;
				if(rsm.getIsReplied() == false) {
					if(bloomFilter.check(rsm.getToUserId().getBytes())) {
						UTPair pair = followeeProfileDB.getUTPair(rsm.getToUserId());
						if(pair != null) {
							if(rsm.getTime() < pair.time) {
								rsm.setIsReplied(true);
								rsm.setAcHops(rsm.getHops());
								rsm.setAcTime(Calendar.getInstance().getTimeInMillis()-rsm.getSTime());
								message.setMessage(rsm);
							} else {
								System.out.println("Don't ask me. I'm also an out-of-date guy!");
							}
						}
					}
				}
			}
			
			if(mesg instanceof CuckooPiggybackMsg) {
				CuckooPiggybackMsg msg = (CuckooPiggybackMsg) mesg;
				
				if(msg.getFromUserId().compareTo(userId) == 0) { //message from myself or to myself
					return null;
				}
				
				if(msg.getToUserId().compareTo(userId) != 0) {
					trafficRecorder.record(msg, Calendar.getInstance().getTimeInMillis(), "f", "i", msg.getFromUserId(), msg.getToUserId(), "n");
				}
				//System.out.println("----Node " + profile.getUserName() + " [" + userId + "] forward----: " + msg.toString());
				monitorPassingMsg(msg);
			}

			return null;
		}
	}
	
	public class WriteStatusWorkTask implements Executable<Object, Exception> {
		private String text;
		
		public WriteStatusWorkTask(String t) {
			super();
			text = t;
		}
		
		/*
		public WriteStatusWorkRequest(Continuation<Object,java.lang.Exception> c, SelectorManager sm, String t) {
			super(c, sm);
			text = t;
		}
		*/
		
		public Object execute() throws Exception {
			writeStatus(text);
			return null;
		}
	}
	
	public class Tweet extends Thread {
		
		public void run() {
			//if(userId.compareTo("96951800") == 0) {
			//if(followerProfileDB.getNumOfEntries() != 0) {
			try {
				BufferedReader tis = new BufferedReader(new FileReader("xml/"+userId +"/"+userId+"_event"));
				String eventLine;
				
				long interval = 0;
				int event = 0;
				////////////////////real code here///////////////////////
				while((eventLine = tis.readLine()) != null) {
					eventLine = eventLine.trim();
					//System.out.println(tweetLine);
					if(eventLine.compareTo("") != 0) {
						event = Integer.valueOf(eventLine.substring(0, 1)); //here the event is only one digit
						if(event == AppPar.SESSION_START) {
							interval = Long.valueOf(eventLine.substring(2, eventLine.length()));
							sleep(interval);
							//System.out.println("[" + userId + "]: " + Calendar.getInstance().getTime() + " join.");
							if(UPLOAD_OR_NOT == true) {
								fetchTweetFromServer();
							}
							bootstrap();
							sayHello();
							
						} else if(event == AppPar.GOSSIPPPPP) {
							String substr = eventLine.substring(2, eventLine.length());
							int idx = substr.indexOf(" ");
							interval = Long.valueOf(substr.substring(0, idx));
							String twtext = substr.substring(idx+1, substr.length());
							sleep(interval);
							//System.out.println("[" + userId + "]: " + Calendar.getInstance().getTime() + " " + twtext);
							write(twtext);
							
						} else if(event == AppPar.SESSION_END) {
							interval = Long.valueOf(eventLine.substring(2, eventLine.length()));
							sleep(interval);
							//System.out.println("[" + userId + "]: " + Calendar.getInstance().getTime() + " quit.");
							quit();
						} else if(event == AppPar.POST_TWEET) {
							//do nothing in the gossip experiment
							
							//String substr = eventLine.substring(2, eventLine.length());
							//int idx = substr.indexOf(" ");
							//interval = Long.valueOf(substr.substring(0, idx));
							//String twtext = substr.substring(idx+1, substr.length());
							long interval2 = 2*24*3600*1000;
							sleep(interval2);
							//System.out.println("[" + userId + "]: " + Calendar.getInstance().getTime() + " " + twtext);
							//write(twtext);
						}
					}
				}
				
				/////////////////////test codes/////////////////////////
				/*String tweetLine;
				interval = (int) (240000.0 * Math.random()); 
				while((tweetLine = tis.readLine()) != null) {
					tweetLine = tweetLine.trim();
					sleep(interval);
					System.out.println("[" + userId + "]:" + Calendar.getInstance().getTime() + tweetLine);
					write(tweetLine);
				}*/
				////////////////////////////////////////////////////////
				die();
				tis.close();
				
			} catch(IOException ioe) {
				ioe.printStackTrace();
			} catch (InterruptedException ie) {
				ie.printStackTrace();
			}
		}
	}
}
